// (c) 2000 Richard Grimes, iDevResource.com Ltd
// you are free to use this code as long as your copyright message
// acknowledges iDevResource.com Ltd

#ifndef __BUCKET_FILE_H
#define __BUCKET_FILE_H

#define BUF_INC 256

// this class gives access to a file
class CFile
{
public:
   // uninitialized file object
   CFile() : m_hFile(INVALID_HANDLE_VALUE)
   {}
   // opens a file with the specified access
   CFile(LPCTSTR strFile, DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE)
       : m_hFile(INVALID_HANDLE_VALUE)
   {
      Open(strFile, dwDesiredAccess);
   }
   // closes the file
   virtual ~CFile()
   {
      Close();
   }
   // opens an existing file
   BOOL Open(LPCTSTR strFile, DWORD dwDesiredAccess = GENERIC_READ)
   {
      if (m_hFile != INVALID_HANDLE_VALUE)
         CloseHandle(m_hFile);
      m_hFile = CreateFile(strFile, dwDesiredAccess, 0, NULL, OPEN_ALWAYS, 0, NULL);
      // move the file pointer to the start
      if (m_hFile != INVALID_HANDLE_VALUE)
         ResetStart();
      return (m_hFile != INVALID_HANDLE_VALUE);
   }
   // creates a new file
   BOOL Create(LPCTSTR strFile, DWORD dwDesiredAccess = GENERIC_WRITE)
   {
      if (m_hFile != INVALID_HANDLE_VALUE)
         CloseHandle(m_hFile);
      m_hFile = CreateFile(strFile, dwDesiredAccess, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
      return (m_hFile != INVALID_HANDLE_VALUE);
   }
   // closes a file
   BOOL Close()
   {
      if (m_hFile != INVALID_HANDLE_VALUE)
      {
         BOOL b = CloseHandle(m_hFile);
         m_hFile = INVALID_HANDLE_VALUE;
         return b;
      }
      return FALSE;
   }
   // moves the file pointer to the beginning of the file
   BOOL ResetStart()
   {
      if (m_hFile != INVALID_HANDLE_VALUE)
         return SetFilePointer(m_hFile, 0, NULL, FILE_BEGIN) != NULL;
      return FALSE;
   }
   // moves the file pointer to the end of the file
   BOOL ResetEnd()
   {
      if (m_hFile != INVALID_HANDLE_VALUE)
         return SetFilePointer(m_hFile, 0, NULL, FILE_END) != NULL;
      return FALSE;
   }
   // reads an ANSI string
   DWORD ReadString(LPSTR str, DWORD dwSize)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(str != NULL);
      if (dwSize == 0)
         return 0;
      DWORD dwRead;
      DWORD dwReturned = 0;
      LPSTR ptr = str;
      while (ReadFile(m_hFile, ptr, dwSize, &dwRead, NULL))
      {
         ptr += dwRead;
         dwReturned += dwRead;
         dwSize -= dwRead;
         if (dwSize == 0)
            break;
      }
      str[dwReturned] = 0;
      return dwReturned;
   }
   // reads a UNICODE string
   DWORD ReadString(LPWSTR str, DWORD dwSize)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(str != NULL);
      if (dwSize == 0)
         return 0;
      DWORD dwRead;
      DWORD dwReturned = 0;
      dwSize *= 2;
      LPWSTR ptr = str;
      while (ReadFile(m_hFile, ptr, dwSize, &dwRead, NULL))
      {
         ptr += dwRead/2;
         dwReturned += dwRead/2;
         dwSize -= dwRead;
         if (dwSize == 0)
            break;
      }
      str[dwReturned] = 0;
      return dwReturned;
   }
   // read to the next \r\n, this will return a buffer allocated with new
   LPSTR ReadLineA()
   {
      DWORD dwRead;
      DynamicString<CHAR> str(-1);
      int pos = 0;
      while (true)
      {
         // read characters a byte at a time, and check for \r\n
         if (!ReadFile(m_hFile, (LPBYTE)(str + pos), 1, &dwRead, NULL) || dwRead == 0)
         {
            str[pos] = 0;
            if (pos == 0)
            {
               // error, back out
               return NULL;
            }
            return str.Detach();
         }
         if (str[pos] == '\r')
         {
            // almost there
            pos ++;
            if (pos > str.m_size)
            {
               // allocate a new buffer if required
               str.IncreaseBy(BUF_INC);
            }
            if (!ReadFile(m_hFile, (LPBYTE)(str + pos), 1, &dwRead, NULL) 
               || dwRead == 0 || str[pos] == '\n')
            {
               // return the string
               str[pos-1] = 0;
               return str.Detach();
            }
         }
         pos++;
         if (pos > str.m_size)
         {
            // allocate the new buffer if required
            str.IncreaseBy(BUF_INC);
         }
      }
      str[pos-1] = 0;
      return str.Detach();
   }
   // read to the next \r\n, this will return a buffer allocated with new
   LPWSTR ReadLineW()
   {
      DWORD dwRead;
      DynamicString<WCHAR> str(-1);
      int pos = 0;
      while (true)
      {
         // read characters a byte at a time, and check for \r\n
         if (!ReadFile(m_hFile, (LPBYTE)(str + pos), 2, &dwRead, NULL) || dwRead == 0)
         {
            str[pos] = 0;
            if (pos == 0)
            {
               // error, back out
               return NULL;
            }
            return str.Detach();
         }
         if (str[pos] == L'\r')
         {
            // almost there
            pos ++;
            if (pos % BUF_INC == 0)
            {
               // allocate a new buffer if required
               str.IncreaseBy(pos);
            }
            if (!ReadFile(m_hFile, (LPBYTE)(str + pos), 2, &dwRead, NULL) || dwRead == 0 || str[pos] == L'\n')
            {
               // return the string
               str[pos-1] = 0;
               return str.Detach();
            }
         }
         pos++;
         if (pos % BUF_INC == 0)
         {
            // allocate the new buffer if required
            str.IncreaseBy(pos);
         }
      }
      str[pos-1] = 0;
      return str.Detach();
   }
   // write an ANSI string to the file
   DWORD WriteString(LPCSTR str)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(str != NULL);
      DWORD dwWritten;
      DWORD dwSize = lstrlenA(str);
      if (dwSize == 0)
         return 0;
      DWORD dwReturned = 0;
      LPCSTR ptr = str;
      while (WriteFile(m_hFile, ptr, dwSize, &dwWritten, NULL))
      {
         ptr += dwWritten;
         dwSize -= dwWritten;
         dwReturned += dwWritten;
         if (dwSize == 0)
            break;
      }
      return dwReturned;
   }
   // write a UNICODE string to the file
   DWORD WriteString(LPCWSTR str)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(str != NULL);
      DWORD dwWritten;
      DWORD dwSize = wcslen(str) * 2;
      if (dwSize == 0)
         return 0;
      DWORD dwReturned = 0;
      LPCWSTR ptr = str;
      while (WriteFile(m_hFile, ptr, dwSize, &dwWritten, NULL))
      {
         ptr += dwWritten/2;
         dwSize -= dwWritten;
         dwReturned += dwWritten/2;
         if (dwSize == 0)
            break;
      }
      return dwReturned;
   }
   // read data directly from the file
   DWORD ReadRaw(LPBYTE pb, DWORD dwSize)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(pb != NULL);
      if (dwSize == 0)
         return 0;
      DWORD dwRead;
      DWORD dwReturned = 0;
      while (ReadFile(m_hFile, pb, dwSize, &dwRead, NULL))
      {
         pb += dwRead;
         dwReturned += dwRead;
         dwSize -= dwRead;
         if (dwSize == 0 || dwRead == 0)
            break;
      }
      return dwReturned;
   }
   // write data directly to the file
   DWORD WriteRaw(LPBYTE pb, DWORD dwSize)
   {
      if (m_hFile == INVALID_HANDLE_VALUE)
         return 0;
      ATLASSERT(pb != NULL);
      DWORD dwWritten;
      DWORD dwReturned = 0;
      while (WriteFile(m_hFile, pb, dwSize, &dwWritten, NULL))
      {
         pb += dwWritten;
         dwSize -= dwWritten;
         dwReturned += dwWritten;
         if (dwSize == 0)
            break;
      }
      return dwReturned;
   }
   // determine if a fle is open (or if an open operation was successful)
   BOOL IsOpen()
   {
      return (m_hFile != INVALID_HANDLE_VALUE);
   }
   // attach to an existing handle taking ownership
   void Attach(HANDLE hHandle)
   {
      Close();
      m_hFile = hHandle;
   }   
   // give up the handle to the caller
   HANDLE Detach()
   {
      HANDLE h = m_hFile;
      m_hFile = INVALID_HANDLE_VALUE;
      return h;
   }
   // duplicates the handle so the callers is still valid
   BOOL AttachNoOwnership(HANDLE hHandle)
   {
      Close();   
      if (!DuplicateHandle(GetCurrentProcess(), hHandle, GetCurrentProcess(), 
                           &m_hFile, 0, TRUE, DUPLICATE_SAME_ACCESS))
      {
         return FALSE;
      }
      return TRUE;
   }
   // conversion operator
   operator HANDLE()
   {
      return m_hFile;
   }
private:
   HANDLE m_hFile;
protected:
   // utility class that allocates the string memory from the heap
   template<typename T>
   class DynamicString
   {
   public:
      // uninitialized buffer
      DynamicString() : m_size(0), m_data(NULL)
      {
      }
      // buffer with a specified size
      DynamicString(int size) : m_size(size)
      {
         if (size == -1) m_size = BUF_INC;
         m_data = new T[m_size];
      }
      // create the buffer based on a string
      DynamicString(const T* str)
      {
         if (sizeof(T) == sizeof(CHAR))
         {
            m_size = strlen((LPCSTR)str) + 1;
            m_data = new T[m_size];
            strcpy((LPSTR)m_data, (LPCSTR)str);
         }
         else
         {
            m_size = wcslen((LPCWSTR)str) + 1;
            m_data = new T[m_size];
            wcscpy((LPWSTR)m_data, (LPCWSTR)str);
         }
      }
      // detach the buffer
      T* Detach()
      {
         T* temp = m_data;
         m_data = NULL;
         return temp;
      }
      // attach a buffer
      T* Attach(T* str)
      {
         if (str != NULL)
         {
            if (sizeof(T) == sizeof(CHAR))
               m_size = strlen((LPCSTR)str) + 1;
            else
               m_size = wcslen((LPCWSTR)str) + 1;
         }
         else
            m_size = 0;
         if (m_data) delete [] m_data;
         m_data = str;
         return m_data;
      }
      // get access to the raw buffer
      operator T*()
      {
         return m_data;
      }
      // get access to the raw buffer
      T* GetBuffer()
      {
         return m_data;
      }
      // clean up the data
      ~DynamicString()
      {
         if (m_data) delete [] m_data;
      }
      // get access to individual characters in the buffer
      T& operator[](int i)
      {
         return m_data[i];
      }
      // change the size of the buffer
      T* IncreaseBy(int size)
      {
         m_size += size;
         T* lpTemp = new T[m_size];
         CopyMemory(lpTemp, m_data, m_size * sizeof(T));
         delete [] m_data;
         return (m_data = lpTemp);
      }
      T* m_data;
      int m_size;
   };
};

// note that these identifiers are ANSI
#define START_PROCESS "START_PROCESS"
#define END_PROCESS "END_PROCESS"

// this is a class that writes data from a bucket to a file
class CBucketFile : public CFile
{
public:
   // create the object empty with a type and version
   CBucketFile(LPCSTR strID, DWORD dwVer) : m_strID(strID), m_dwVer(dwVer)
   {      
   }
   // creates a new file
   BOOL Create(LPCTSTR strFile, DWORD dwDesiredAccess = GENERIC_WRITE)
   {
      // create the file and write our identifier
      BOOL b = CFile::Create(strFile, dwDesiredAccess);
      if (b)
      {
         WriteString(m_strID);
         WriteString("\r\n");
         DynamicString<CHAR> str(10);
         sprintf(str, "%ld\r\n", m_dwVer);
         WriteString(str);
      }
      return b;
   }
   // opens an existing bucket file
   BOOL Open(LPCTSTR strFile, DWORD dwDesiredAccess = GENERIC_READ)
   {
      BOOL b = CFile::Open(strFile, dwDesiredAccess);
      if (b)
      {
         // we need to check that we can read the file
         DynamicString<CHAR> strID;
         if (strID.Attach(ReadLineA()) != NULL)
         {
            // check the file ID
            if (strcmp(strID, m_strID) == 0)
            {
               // check that the version is less than or equal to one 
               // we support
               DynamicString<CHAR> strVer;
               if (strVer.Attach(ReadLineA()) != NULL)
               {
                  DWORD dwVer = (DWORD)atol(strVer);
                  if (dwVer <= m_dwVer)
                     return TRUE;
               }
            }
         }
         Close();
         return FALSE;
      }
      return b;
   }
   // returns TRUE if the operation completed successfully
   BOOL FillBucket(CBucket* pBucket)
   {
      while (true)
      {
         DynamicString<CHAR> strLine;
         strLine.Attach(ReadLineA());
         if (strLine == NULL)
            return FALSE;
         if (strcmp(strLine, START_PROCESS) == 0)
         {
            continue;
         }
         if (strcmp(strLine, END_PROCESS) == 0)
         {
            break;
         }
         if (strncmp(strLine, "Process", 7) == 0)
         {
            strLine[7] = 0;
            pBucket->m_pid = atol(strLine + 8);
            continue;
         }
         pBucket->Add(strLine);
      }
      return TRUE;
   }
   BOOL SaveBucket(CBucket* pBucket)
   {
      ATLASSERT(pBucket != NULL);
      if (!ResetEnd())
         return FALSE;
      WriteString(START_PROCESS);
      WriteString("\r\n");
      CHAR str[50];
      sprintf(str, "Process %ld\r\n", pBucket->m_pid);
      WriteString(str);

      CBucket::iterator it;
      for (it = pBucket->begin(); it < pBucket->end(); it++)
      {
         CBucket::BucketData& bd = (*it);
         sprintf(str, "%04d-%02d-%02d %02d:%02d:%02d.%03d$ ",
            bd.st.wYear, bd.st.wMonth, bd.st.wDay,
            bd.st.wHour, bd.st.wMinute, bd.st.wSecond, bd.st.wMilliseconds);
         WriteString(str);
         WriteString(bd.str);
         WriteString("\r\n");
      }

      WriteString(END_PROCESS);
      WriteString("\r\n");
      return TRUE;
   }
private:
   DynamicString<CHAR> m_strID;
   DWORD m_dwVer;
};

#endif  //__BUCKET_FILE_H
