// (c) 2000 Richard Grimes, iDevResource.com Ltd
// you are free to use this code as long as your copyright message
// acknowledges iDevResource.com Ltd

#include "StdAfx.h"
#include "DebugInfoView.h"
#include "BucketFile.h"

// this is just to make resource management simpler
class CVars
{
public:
   // for convenience this gives access to the data in the MMF written to
   // by OutputDebugString()
   typedef struct _DebugData
   {
      DWORD pid;
      char str[1];
   } DebugData, *LPDEBUGDATA;
   // holds descriptions of exceptions
   typedef struct _ExceptData
   {
      DWORD code;
      LPCTSTR desc;
   } ExceptionData;

   CVars() : 
      hAck(NULL), hReady(NULL), hMap(NULL), pData(NULL), str(NULL)
   {}
   ~CVars()
   {
      if (hAck) CloseHandle(hAck); 
      if (hReady) CloseHandle(hReady); 
      if (hMap) CloseHandle(hMap); 
      if (pData) UnmapViewOfFile(pData);
      if (str) delete [] str;
   }
   DWORD RetVal()
   {
      return GetLastError(); 
   }
   // gives access to an array with the commonly thrown exceptiuons
   static ExceptionData* GetExceptionData()
   {
      static ExceptionData arr[] = 
      {
         {EXCEPTION_ACCESS_VIOLATION, _T("Access violation")}, 
         {EXCEPTION_ARRAY_BOUNDS_EXCEEDED, _T("Accessed an array element out of bounds")},
         {EXCEPTION_BREAKPOINT, _T("Breakpoint")},
         {EXCEPTION_DATATYPE_MISALIGNMENT, _T("Tried to read or write data that is misaligned")},
         {EXCEPTION_FLT_DENORMAL_OPERAND, _T("Floating point number is too small to be represent as a standard floating-point value.")},
         {EXCEPTION_FLT_DIVIDE_BY_ZERO, _T("Floating point divide by zero")},
         {EXCEPTION_FLT_INEXACT_RESULT, _T("The result cannot be represented exactly as a decimal fraction.")},
         {EXCEPTION_FLT_INVALID_OPERATION, _T("Floating point invalid operation")},
         {EXCEPTION_FLT_OVERFLOW, _T("Floating point overflow")},
         {EXCEPTION_FLT_STACK_CHECK, _T("Floating point stack overflow or underflow")},
         {EXCEPTION_FLT_UNDERFLOW, _T("Floating point underflow")},
         {EXCEPTION_ILLEGAL_INSTRUCTION, _T("Illegal instruction.")},
         {EXCEPTION_IN_PAGE_ERROR, _T("Page error")},
         {EXCEPTION_INT_DIVIDE_BY_ZERO, _T("Divide by zero")},
         {EXCEPTION_INT_OVERFLOW, _T("Integer overflow")},
         {EXCEPTION_INVALID_DISPOSITION, _T("Exception handler invalid disposition")},
         {EXCEPTION_NONCONTINUABLE_EXCEPTION, _T("Noncontinuable execution")},
         {EXCEPTION_PRIV_INSTRUCTION, _T("Operation is not allowed in the current machine mode")},
         {EXCEPTION_SINGLE_STEP, _T("Single step")},
         {EXCEPTION_STACK_OVERFLOW, _T("Stack overflow")},
         {0, NULL}
      };

      return arr;
   }
   HANDLE hAck;
   HANDLE hReady;
   HANDLE hMap;
   LPDEBUGDATA pData;
   LPSTR str;
};

// the process is starting, so we must initialize it
LRESULT CDebugInfoView::OnInitDialog(HWND, LPARAM)
{
   // set up the tree view
   m_imageList.Create(MAKEINTRESOURCE(IDB_TREEIMG),  16, FALSE, CLR_NONE);

	CTreeViewCtrl tv;
   tv = GetDlgItem(IDC_TREEDATA);
   tv.SetImageList(m_imageList, TVSIL_NORMAL);

   // center the dialog on the screen
	CenterWindow();

	// set icons
	HICON hIcon = (HICON)::LoadImage(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDR_MAINFRAME), 
		IMAGE_ICON, ::GetSystemMetrics(SM_CXICON), ::GetSystemMetrics(SM_CYICON), LR_DEFAULTCOLOR);
	SetIcon(hIcon, TRUE);
	HICON hIconSmall = (HICON)::LoadImage(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDR_MAINFRAME), 
		IMAGE_ICON, ::GetSystemMetrics(SM_CXSMICON), ::GetSystemMetrics(SM_CYSMICON), LR_DEFAULTCOLOR);
	SetIcon(hIconSmall, FALSE);

   // now set up the event that is used to kill the threads
   m_hDie = CreateEvent(NULL, TRUE, FALSE, NULL);
   DWORD dwThreadId;

   // create the thread that will get messages from the debug stream
   m_hThread = CreateThread(NULL, 0, PollDebugStream, this, 0, &dwThreadId);

   // see if the user passed the name of a process on the command line 
   // this will happen if this has been registered as the system debugger
   m_strCommandLine = GetCommandLine();
   if (m_strCommandLine.GetLength() > 0)
   {
      // do we have the h switch? if so hide me
      if (GetSwitch(_T("h"), NULL))
      {
         ShowWindow(SW_HIDE);
      }
      // was this started to debug a process? 
      if (GetSwitch(_T("p"), NULL))
      {
         // attach to a process
         EXEData* pData = new EXEData;
         pData->pThis = this;      
         pData->strEXE[0] = 0;

         HANDLE hThread;
         DWORD dwThreadId;
         hThread = CreateThread(NULL, 0, PollDebuggerEvents, 
            pData, 0, &dwThreadId);
         if (hThread == NULL)
         {
            delete pData;
         }

         // PollDebuggerEvents will add the thread to the vector
         // so we can free the handle
         CloseHandle(hThread);
      }
   }
	return TRUE;
}

// this is used to make sure that the process cannot be sized to nothing
LRESULT CDebugInfoView::OnGetMinMax(LPMINMAXINFO lpMinMax)
{
   CRect rect;
   GetWindowRect(rect);

   CTreeViewCtrl tv = GetDlgItem(IDC_TREEDATA);
   CRect rectTV;
   tv.GetWindowRect(&rectTV);
   CStatic stat = GetDlgItem(IDC_PROCESSPATH);
   CRect rectStat;
   stat.GetWindowRect(rectStat);

   int minx, miny;
   minx = rectTV.Width();
   miny = rectStat.Height();

   // get the size of the border
   int cxBorder = GetSystemMetrics(SM_CXFRAME);
   int cyBorder = GetSystemMetrics(SM_CYFRAME);

   // frame border
   minx += cxBorder * 2;
   miny += cyBorder * 2;

   // caption and menu
   int cTop = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU);
   miny += cTop;

   // position of the window when maximized
   lpMinMax->ptMaxPosition.x = 0;
   lpMinMax->ptMaxPosition.y = 0;

   // size of the window when maximized
   lpMinMax->ptMaxSize.x = GetSystemMetrics(SM_CXSCREEN);
   lpMinMax->ptMaxSize.y = GetSystemMetrics(SM_CYSCREEN);

   // max tracking size of the window
   lpMinMax->ptMaxTrackSize.x = GetSystemMetrics(SM_CXSCREEN);
   lpMinMax->ptMaxTrackSize.y = GetSystemMetrics(SM_CYSCREEN);

   // min tracking size of the window
   lpMinMax->ptMinTrackSize.x = minx;
   lpMinMax->ptMinTrackSize.y = miny;
   return 0;
}

// when the main window is resized we want to resize and position the
// controls on the dialog
LRESULT CDebugInfoView::OnSize(UINT flags, CSize size)
{
   CRect rect;
   GetWindowRect(rect);

   CTreeViewCtrl tv = GetDlgItem(IDC_TREEDATA);
   CRect rectTV;
   tv.GetWindowRect(&rectTV);
   CStatic stat = GetDlgItem(IDC_PROCESSPATH);
   CRect rectStat;
   stat.GetWindowRect(rectStat);

   // get the size of the border, this is the space between controls
   int cxBorder = GetSystemMetrics(SM_CXFRAME);
   int cyBorder = GetSystemMetrics(SM_CYFRAME);

   // resize the controls
   CListBox lb = GetDlgItem(IDC_MSGS);
   tv.SetWindowPos(0, 
      0, 0, 
      rectTV.Width(), rect.Height(),
      SWP_NOZORDER);
   stat.SetWindowPos(0, 
      rectTV.Width() + cxBorder, 0,
      rect.Width() - rectTV.Width() - cxBorder, rectStat.Height(),
      SWP_NOZORDER);
   lb.SetWindowPos(0, 
      rectTV.Width() + cxBorder, rectStat.Height() + cyBorder,
      rect.Width() - rectTV.Width() - cxBorder, rect.Height() - rectStat.Height() - cyBorder,
      SWP_NOZORDER);
	return 0;
}

// the user has clicked on the tree view, so lets put the data in the list box
LRESULT CDebugInfoView::OnSelectedPID(int idCtrl, LPNMHDR pnmh, BOOL& bHandled)
{
   LPNMTREEVIEW ptv = reinterpret_cast<LPNMTREEVIEW>(pnmh);
   CBucket* pBucket = reinterpret_cast<CBucket*>(ptv->itemNew.lParam);

   // a bucket holds the debug string data for a process
   if (pBucket)
   {
      // get the pid that the user clicked on
      CTreeViewCtrl tv;
      tv = GetDlgItem(IDC_TREEDATA);
      
      // get the selected item
      HTREEITEM hItem = NULL;
      hItem = tv.GetNextItem(TVGN_ROOT, TVGN_CARET);

      // now get its data
      TCHAR szItem[20];
      TVITEM item;
      item.hItem = hItem;
      item.mask = TVIF_TEXT | TVIF_HANDLE;
      item.pszText = szItem;
      item.cchTextMax = sizeof(szItem);
      tv.GetItem(&item);
      DWORD dwPid = _ttoi(szItem);

      // write the process ID to the static control
      TCHAR szProcess[25];
      lstrcpy(szProcess, _T("PID: "));
      lstrcat(szProcess, szItem);

      CStatic wnd;
      wnd = GetDlgItem(IDC_PROCESSPATH);
      wnd.SetWindowText(szProcess);

      // look in the array to see if we have attached to the process
      // if so, put the full path in the static control
      std::vector<ThreadData>::iterator iter;
      m_threadArrayCS.Lock();
      for (iter = m_hDebugThreads.begin(); iter < m_hDebugThreads.end(); iter++)
      {
         ThreadData& td = (*iter);
         if (td.pid == dwPid)
         {
            wnd.SetWindowText(td.str);
            break;
         }
      }
      m_threadArrayCS.Unlock();

      // create a buffer for the data in this bucket
      LPSTR str = new CHAR[pBucket->GetMax() + 21];
      CListBox lb;
      lb = GetDlgItem(IDC_MSGS);
      lb.ResetContent();

      // now go through this bucket and get all the messages in it
      CBucket::iterator it;
      for (it = pBucket->begin(); it < pBucket->end(); it++)
      {
         // format the data in a string
         CBucket::BucketData& bd = *it;
         wsprintf(str, _T("%04d-%02d-%02d %02d:%02d:%02d %s"),
            bd.st.wYear, bd.st.wMonth, bd.st.wDay,
            bd.st.wHour, bd.st.wMinute, bd.st.wSecond, bd.str);

         int index;
         // add the string to the list box and make sure that the
         // scroll bars are sized accordingly
         index = lb.AddString(str);
         SizeListbox(lb, str);
         lb.SetItemData(index, reinterpret_cast<DWORD>(pBucket));
      }

      delete [] str;
   }
	return 0;
}

// This is the routine that accepts data from the debugging threads, it will
// add the message to a 'bucket' associated with the treeview item for the
// process that generated the message.
// lParam is the string of the message
// wParam is the PID of the process
LRESULT CDebugInfoView::OnNewData(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
   LPTSTR str = reinterpret_cast<LPTSTR>(lParam);
   TCHAR szPID[20];
   wsprintf(szPID, _T("%ld"), wParam);

   // see if this PID is in the treeview
	CTreeViewCtrl tv;
   tv = GetDlgItem(IDC_TREEDATA);
   UINT count = tv.GetCount();

   HTREEITEM hItem;
   HTREEITEM hNewItem = NULL;
   hItem = tv.GetNextItem(TVGN_ROOT, 0);
   while (count--)
   {
      // get the item's data
      TCHAR szItem[MAX_PATH];
      TVITEM item;
      item.hItem = hItem;
      item.mask = TVIF_TEXT | TVIF_HANDLE;
      item.pszText = szItem;
      item.cchTextMax = sizeof(szItem);
      tv.GetItem(&item);
      // check to see if it is for the process that we have got a message for
      if (lstrcmp(szItem, szPID) == 0)
      {
         hNewItem = hItem;
         break;
      }
      // no, continue and look at the next tree view item
      hItem = tv.GetNextItem(hItem, TVGN_NEXT);
   }

   CBucket* pBucket = NULL;

   // if the PID has not been seen before then we must add a 
   // new branch to the tree view
   if (hNewItem == NULL)
   {
      pBucket = new CBucket(wParam);
      TVINSERTSTRUCT tvis = {TVI_ROOT, TVI_LAST, {TVIF_TEXT|TVIF_PARAM|TVIF_IMAGE, 0, 0, 0, szPID, 0, 0, 0, 0, reinterpret_cast<LPARAM>(pBucket)}};
      tv.InsertItem(&tvis);
   }
   else
   // we have a message for a process that we have seen before
   {
      TVITEM tvi = {TVIF_PARAM | TVIF_HANDLE, hNewItem, 0, 0, NULL, 0, 0, 0, 0, 0};
      tv.GetItem(&tvi);
      pBucket = reinterpret_cast<CBucket*>(tvi.lParam);
   }

   // since we have added an item we must enable the save menu item and clear item
   m_pParent->UIEnable(ID_FILE_SAVERESULTS, TRUE);
   m_pParent->UIEnable(ID_VIEW_CLEARALL, TRUE);
   // update the status bar, send this so that the string is still valid
   // when the message is handled
   CString strPid;
   strPid.Format(_T("Last message from %ld"), wParam);
   ::SendMessage(GetParent(), WM_UPDATESB, reinterpret_cast<WPARAM>(strPid.operator LPCTSTR()), 0);

   // we need to make sure that the string is formatted correctly
   // this will remove tabs and new lines
   str = SplitString(str, pBucket);
   if (str)
   {
      // add the data to the bucket
      pBucket->Add(str);
      delete [] str;
   }
   return 0;
}

// go through string and change any tabs to spaces and /r/n to new entries
LPTSTR CDebugInfoView::SplitString(LPTSTR str, CBucket* pBucket)
{
   LPTSTR ptr = str;
   while (*ptr != 0)
   {
      if (*ptr == _T('\r') || *ptr == _T('\n'))
      {
         *ptr = 0;
         if (ptr[1] == _T('\n'))
            ptr += 2;
         else
            ptr++;

         // add the data
         if (lstrlen(str) > 0)
         {
            pBucket->Add(str);
         }
         if (*ptr != 0)
         {
            LPSTR newstr = new char[lstrlen(ptr) + 1];
            lstrcpy(newstr, ptr);
            delete [] str;
            ptr = str = newstr;
            continue;
         }
         else
         {
            delete [] str;
            str = 0;
            break;
         }
      }
      if (*ptr == _T('\t'))
      {
         int len = lstrlen(str) + TAB + 1;
         LPTSTR newstr = new TCHAR[len];
         if (ptr > str)
            memcpy(newstr, str, ptr - str);
         for (int x = 0; x < TAB; x++)
            newstr[ptr - str + x] = _T(' ');
         int pos = ptr - str + 1;
         memcpy(newstr + (ptr - str) + TAB, str + pos, lstrlen(str) - (ptr - str));
         delete [] str;
         str = newstr;
         ptr = str;
      }
      ptr++;
   }
   return str;
}

// determine how big the horiz scroll bar should be
void CDebugInfoView::SizeListbox(HWND hWnd, LPTSTR str)
{
   CListBox wnd(hWnd);
   long width;
   width = wnd.GetWindowLong(GWL_USERDATA);

   HDC hdc;
   hdc = wnd.GetDC();
   SIZE size;
   GetTextExtentPoint32(hdc, str, lstrlen(str), &size);

   if (size.cx > width)
      width = size.cx;

   wnd.SetHorizontalExtent(width);
   wnd.ReleaseDC(hdc);
   wnd.SetWindowLong(GWL_USERDATA, width);
}

// Static method, used as a thread function to catch data that was sent to
// the debug stream. Much of this code was taken from the DBMon sample 
// in the platform SDK
DWORD WINAPI CDebugInfoView::PollDebugStream(LPVOID pv)
{
   CDebugInfoView* pThis = static_cast<CDebugInfoView*>(pv);
   // this is used to manage resources
   CVars vars;

   // first create a NULL DACL, so that we can access the MMF
   SECURITY_ATTRIBUTES sa;
   SECURITY_DESCRIPTOR sd;

   sa.nLength = sizeof(SECURITY_ATTRIBUTES);
   sa.bInheritHandle = TRUE;
   sa.lpSecurityDescriptor = &sd;

   if (!InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION))
      return vars.RetVal();
   if (!SetSecurityDescriptorDacl(&sd, TRUE, (PACL)NULL, FALSE)) 
      return vars.RetVal();

   // access the events that control access to the MMF
   vars.hAck = CreateEvent(&sa, FALSE, FALSE, _T("DBWIN_BUFFER_READY"));
   if (!vars.hAck)
      return vars.RetVal();
   vars.hReady = CreateEvent(&sa, FALSE, FALSE, _T("DBWIN_DATA_READY"));
   if (!vars.hReady)
      return vars.RetVal();

   // now access the MMF
   vars.hMap = CreateFileMapping((HANDLE)-1, &sa, PAGE_READWRITE, 0, 4096, _T("DBWIN_BUFFER"));
   if (!vars.hMap)
      return vars.RetVal();
   vars.pData = reinterpret_cast<CVars::LPDEBUGDATA>(MapViewOfFile(vars.hMap, FILE_MAP_READ, 0, 0, 512));
   if (!vars.pData)
      return vars.RetVal();

   // create a buffer for the data
   SYSTEM_INFO si;
   GetSystemInfo(&si);

   // data is only ANSI
   vars.str = new CHAR[si.dwPageSize];
   vars.str[0] = 0;

   DWORD dwRet;
   // loop until we are told to die
   while(true)
   {
      if (WaitForSingleObject(pThis->m_hDie, pThis->m_dwTimeout) == WAIT_OBJECT_0)
         // bye bye
         break;

      // say that we are free to read from the MMF
      SetEvent(vars.hAck);
      // wait for the MMF to fill
      dwRet = WaitForSingleObject(vars.hReady, TIMEOUT);
      if (dwRet == WAIT_OBJECT_0)
      {
         // get the data, read it in and put it in the buffer
         // until a newline is obtained. The first DWORD is the PID of the
         // process that generated the message
         bool bCR = false;
         strcat(vars.str, vars.pData->str);
         bCR = (vars.str[strlen(vars.str) - 1] == '\n');
         if (bCR)
         {
            // a newline was obtained, so post the message to the 
            // main dialog
            int len = strlen(vars.str) + 1;
            LPTSTR newData = new TCHAR[len];
            // the data in the MMF is ANSI, so if you insist on 
            // compiling for UNICODE, then we will have to convert it
#ifdef UNICODE
            MultiByteToWideChar(CP_ACP, 0, vars.str, -1, newData, len);
#else
            lstrcpy(newData, vars.str);
#endif
            vars.str[0] = 0;
            // use PostMessage as an inter thread mechanism...
            pThis->PostMessage(WM_NEWDATA, vars.pData->pid, reinterpret_cast<LPARAM>(newData));
         }
      }
      else if (dwRet == WAIT_TIMEOUT)
         continue;
   }

   // resources are managed by the vars variable
   return 0;
}

// Attach to a process as a debugger, note that you cannot unattach
// a debugger.
//
// This has two options, either start a process, in which case the name and
// path is passed via the parameter; or attach to a running process. In the
// second case the command line passed to DebugInfo is in this form:
//    /p processid /e eventhandle
// processid is the PID for the running process. This command line is used
// by the system when it attaches a debugger registered in the AeDebug
// registry value. The eventhandle is an event that is used to tell the 
// system that we have successfully attached.
DWORD WINAPI CDebugInfoView::PollDebuggerEvents(LPVOID pv)
{
   // information about the EXE name and where to send data to
   EXEData* pData = static_cast<EXEData*>(pv);
   CDebugInfoView* pThis = static_cast<CDebugInfoView*>(pData->pThis);

   // this holds data about the DLLs loaded by the debugged process
   CSimpleArray<DLLData> arrDLL;

   STARTUPINFO sui;
   PROCESS_INFORMATION pi;
   memset(&sui, 0, sizeof(STARTUPINFO));
   sui.cb = sizeof(STARTUPINFO);
   sui.dwFlags = STARTF_USESHOWWINDOW;
   sui.wShowWindow = SW_SHOWNORMAL;
   memset(&pi, 0, sizeof(PROCESS_INFORMATION));

   // this was started by the system to attach to a process
   if (pData->strEXE[0] == 0)
   {
      if (pThis->m_strCommandLine.GetLength() == 0)
      {
         delete pData;
         return 0;
      }

      // get the process ID from the command line
      TCHAR strValue[10];
      DWORD dwPid = 0;
      if (!pThis->GetSwitch(_T("p"), strValue))
      {
         delete pData;
         return 0;
      }

      LPTSTR lpstrCmd = strValue;

      // convert to a number, take into account that it might be
      // in hex
      if (lpstrCmd[0] == _T('0') && lpstrCmd[1] == _T('x')) 
      {
         // in hex
         lpstrCmd += 2;
         while ((*lpstrCmd >= _T('0') && *lpstrCmd <= _T('9')) 
            ||(*lpstrCmd >= _T('A') && *lpstrCmd <= _T('F')) 
            ||(*lpstrCmd >= _T('a') && *lpstrCmd <= _T('f'))) 
         {
            if ((*lpstrCmd >= _T('0') && *lpstrCmd <= _T('9')))
               dwPid = dwPid * 10 + *lpstrCmd - _T('0');
            else if ((*lpstrCmd >= _T('A') && *lpstrCmd <= _T('F')))
               dwPid = dwPid * 10 + (*lpstrCmd - _T('A') + 10);
            else if ((*lpstrCmd >= _T('a') && *lpstrCmd <= _T('f')))
               dwPid = dwPid * 10 + (*lpstrCmd - _T('a') + 10);

            lpstrCmd++;
         }
      }
      else 
      {
         // decimal
         while (*lpstrCmd >= _T('0') && *lpstrCmd <= _T('9')) 
         {
            dwPid = dwPid * 10 + *lpstrCmd - _T('0');
            lpstrCmd++;
         }
      }

      // get the event handle
      strValue[0] = 0;
      if (!pThis->GetSwitch(_T("e"), strValue))
      {
         delete pData;
         return 0;
      }

      // this is the actual value of an event handle that is used
      // to indicate to the system that DebugInfo has attached
      HANDLE hEvent = 0;
      lpstrCmd = strValue;
      while (*lpstrCmd >= _T('0') && *lpstrCmd <= _T('9')) 
      {
         hEvent = (HANDLE)((DWORD)hEvent * 10 + *lpstrCmd - _T('0'));
         lpstrCmd++;
      }

      // now attach to the running process
      if (!DebugActiveProcess(dwPid))
      {
         delete pData;
         return 0;
      }

      // indicate that we've attached
      SetEvent(hEvent);
      pi.dwProcessId = dwPid;
      pi.hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwPid);
      // not sure about the thread value :-(
   }
   else
   {
      // we have been asked to start a process
      if (!CreateProcess(NULL, pData->strEXE, NULL, NULL, FALSE, DEBUG_ONLY_THIS_PROCESS, 
          NULL, NULL, &sui, &pi))
      {
         pThis->MessageBox(_T("Cannot create process"));
         delete pData;
         return 0;
      }
      else
      {
         // lower the priority of the new process, so that we get priority
         SetThreadPriority(pi.hThread, THREAD_PRIORITY_BELOW_NORMAL);
      }
   }

   // We can get debug messages from multiple processes, so add our data
   // to an array so that the threads can be cleared up at a later stage
   ThreadData td;
   td.hThread = GetCurrentThread();
   td.pid = pi.dwProcessId;
   lstrcpy(td.str, pData->strEXE);
   
   // put information about this thread in the process thread info array
   pThis->m_threadArrayCS.Lock();
   pThis->m_hDebugThreads.push_back(td);
   pThis->m_threadArrayCS.Unlock();

   // now start the debug message loop
   LPBYTE pDebug = NULL;
   while(true)
   {
      // the process is shutting down and this thread has been asked to die
      if (WaitForSingleObject(pThis->m_hDie, pThis->m_dwTimeout) == WAIT_OBJECT_0)
         break;
      DEBUG_EVENT dbevent;
      // check for the debug event that has been passed to us
      if (WaitForDebugEvent(&dbevent, TIMEOUT))
      {
         // this is a message for another process so we cannot handle it
         if (dbevent.dwProcessId != pi.dwProcessId)
         {
            CleanUpEventResources(dbevent);
            ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
            continue;
         }
         if (dbevent.dwDebugEventCode == OUTPUT_DEBUG_STRING_EVENT)
         {
            // the process has generated a debug string
            if (pDebug == NULL)
            {
               // create a buffer for the data
               SYSTEM_INFO si;
               GetSystemInfo(&si);
               pDebug = new BYTE[si.dwPageSize];
               pDebug[0] = 0;
            }

            DWORD dwRead;
            LPVOID newData;
            // it is a UNICODE string
            if (dbevent.u.DebugString.fUnicode)
            {
               // get the string
               int end = wcslen(reinterpret_cast<LPCWSTR>(pDebug));
               ReadProcessMemory(pi.hProcess, 
                  dbevent.u.DebugString.lpDebugStringData, pDebug + end, 
                  dbevent.u.DebugString.nDebugStringLength, &dwRead);
               // see if the string is terminated with a newline
               if (pDebug[end + dwRead - 2] != L'\n')
               {
                  // no, so more data to come
                  CleanUpEventResources(dbevent);
                  ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
                  continue;
               }
               // conditional compilation 
               // if we have been compiled for UNICODE do nothing
#ifdef UNICODE
               newData = reinterpret_cast<LPVOID>(pDebug);
#else          // if we have been compiled for ANSI, convert it
               LPCWSTR wStr = reinterpret_cast<LPCWSTR>(pDebug);
               LPSTR newStr = new CHAR[wcslen(wStr) + 1];
               WideCharToMultiByte(CP_ACP, 0, wStr, -1, newStr, 
                  (dbevent.u.DebugString.nDebugStringLength) / 2 + 1, 0, 0);
               delete [] pDebug;
#endif
            }
            // it is an ANSI string
            else
            {
               int end = strlen(reinterpret_cast<LPCSTR>(pDebug));
               ReadProcessMemory(pi.hProcess, 
                  dbevent.u.DebugString.lpDebugStringData, pDebug + end, 
                  dbevent.u.DebugString.nDebugStringLength, &dwRead);
               // see if it is terminated with a newline
               if (pDebug[end + dwRead - 2] != '\n')
               {
                  // more data to come
                  CleanUpEventResources(dbevent);
                  ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
                  continue;
               }
#ifdef UNICODE // UNICODE build, so convert to ANSI
               LPCSTR aStr = reinterpret_cast<LPCSTR>(pDebug);
               LPWSTR newStr = new WCHAR[strlen(aStr) + 1];
               MultiByteToWideChar(CP_ACP, 0, aStr, -1, newStr, 
                  dbevent.u.DebugString.nDebugStringLength + 1);
               newData = reinterpret_cast<LPVOID>(newStr);
               delete [] pDebug;
#else          // ANSI build so do nothing
               newData = reinterpret_cast<LPVOID>(pDebug);
#endif
            }

            // Phew! Now that we are here, we can store the message 
            pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(newData));
            // the WM_NEWDATA handler will delete the buffer
            pDebug = NULL;
         }
         // was an exception thrown?
         else if (dbevent.dwDebugEventCode == EXCEPTION_DEBUG_EVENT)
         {
            LPTSTR newData = new TCHAR[256];
            if (dbevent.u.Exception.dwFirstChance)
               lstrcpy(newData, _T("First chance exception: "));
            else
               lstrcpy(newData, _T("Exception: "));

            // read the description of the exception
            CVars::ExceptionData* pEx = CVars::GetExceptionData();
            while (pEx->desc != NULL)
            {
               if (pEx->code 
                    == dbevent.u.Exception.ExceptionRecord.ExceptionCode)
               {
                  lstrcat(newData, pEx->desc);
                  break;
               }
               pEx++;
            }
            
            // add the address where it was generated
            lstrcat(newData, _T(" "));
            _stprintf(newData + lstrlen(newData), _T("Address 0x%0x"),
               dbevent.u.Exception.ExceptionRecord.ExceptionAddress);

            pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(newData));
            
            CleanUpEventResources(dbevent);
            ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
            continue;
         }
         // a DLL has been unloaded
         else if (dbevent.dwDebugEventCode == UNLOAD_DLL_DEBUG_EVENT)
         {
            for (int i = 0; i < arrDLL.GetSize(); i++)
            {
               if (arrDLL[i].lpBase == dbevent.u.UnloadDll.lpBaseOfDll)
               {
                  LPTSTR newData = new TCHAR[lstrlen(arrDLL[i].strDLL) + 7];
                  lstrcpy(newData, _T("UNLD: "));
                  lstrcpy(newData + 6, arrDLL[i].strDLL);
                  pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(newData));
                  break;
               }
            }
         }
         // a DLL has been loaded
         else if (dbevent.dwDebugEventCode == LOAD_DLL_DEBUG_EVENT)
         {
            LPVOID pBuf;
            DLLData data;
            data.lpBase = dbevent.u.LoadDll.lpBaseOfDll;

            ReadProcessMemory(pi.hProcess, dbevent.u.LoadDll.lpImageName,
                  &pBuf, sizeof(LPVOID), NULL);
            if (pBuf == NULL)
            {
               CleanUpEventResources(dbevent);
               ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
               continue;
            }

            // get the name of the DLL
            LPBYTE newData = new BYTE[(MAX_PATH + 1) * 2];
            lstrcpy(reinterpret_cast<LPTSTR>(newData), _T("LOAD: "));
            DWORD dwSize = 0;
            if (ReadProcessMemory(pi.hProcess, pBuf,
                  newData + 6*sizeof(TCHAR), MAX_PATH, &dwSize))
            {
               if (dbevent.u.LoadDll.fUnicode)
               {
                  // the process is UNICODE
#ifdef UNICODE
                  // we were compiled for UNICODE so everything is OK, 
                  // we don't need any conversions              
#else
                  // we were compiled for ANSI, so we need to convert
                  LPSTR strData = new char[dwSize + 7];
                  lstrcpy(strData, "LOAD: ");
                  WideCharToMultiByte(CP_ACP, 0, reinterpret_cast<LPCWSTR>(newData + 6), -1, strData + 6, dwSize + 1, NULL, NULL);
                  lstrcpy(data.strDLL, strData + 6);
                  pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(strData));
                  // get rid of the old buffer that we no longer need
                  delete [] newData;
#endif
               }
               else
               {
                  // the process is ANSI build
#ifdef UNICODE
                  // we were compiled for UNICODE so we need to convert
                  LPWSTR strData = new WCHAR[dwSize + 7];
                  lstrcpy(strData, L"LOAD: ");
                  MultiByteToWideChar(CP_ACP, 0, reinterpret_cast<LPSTR>(newData) + 6, -1, strData, dwSize + 7);
                  lstrcpy(data.strDLL, strData + 6);
                  pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(strData));
                  // get rid of the old buffer that we do not need
                  delete [] newData;
#else
                  // don't need to convert
                  lstrcpy(data.strDLL, reinterpret_cast<LPSTR>(newData + 6));
                  pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(newData));
#endif
               }
               // add this array to the list of DLLs that this process uses
               arrDLL.Add(data);
            }
            else
               delete [] newData;
         }
         // the process has decided to go away
         else if (dbevent.dwDebugEventCode == EXIT_PROCESS_DEBUG_EVENT)
         {
            LPTSTR newData = new TCHAR[30];
            wsprintf(newData, _T("Process terminated: %08x"), dbevent.u.ExitProcess.dwExitCode);
            pThis->PostMessage(WM_NEWDATA, dbevent.dwProcessId, reinterpret_cast<LPARAM>(newData));
            CleanUpEventResources(dbevent);
            break;
         }
         CleanUpEventResources(dbevent);
         ContinueDebugEvent(dbevent.dwProcessId, dbevent.dwThreadId, DBG_CONTINUE);
      }  
   }

   // clean up
   if (pi.hProcess)
   {
      TerminateProcess(pi.hProcess, 0);
      CloseHandle(pi.hProcess);
   }
   if (pi.hThread)
      CloseHandle(pi.hThread);

   // clear up information for this thread
   pThis->m_threadArrayCS.Lock();
   std::vector<ThreadData>::iterator it;
   for (it = pThis->m_hDebugThreads.begin(); 
         it < pThis->m_hDebugThreads.end(); it++)
   {
      ThreadData& td = *it;
      if (td.pid == pi.dwProcessId)
      {
         // found it!
         CloseHandle(td.hThread);
         td.hThread = 0;
         pThis->m_hDebugThreads.erase(it);
         break;
      }
   }
   pThis->m_threadArrayCS.Unlock();
   delete pData;

   return 0;
}

// this is used to clean up debugger resources according to the debugger event
void CDebugInfoView::CleanUpEventResources(DEBUG_EVENT dbevent)
{
   switch (dbevent.dwDebugEventCode)
   {
   case CREATE_THREAD_DEBUG_EVENT:
      if (dbevent.u.CreateThread.hThread)
         CloseHandle(dbevent.u.CreateThread.hThread);
      break;
   case CREATE_PROCESS_DEBUG_EVENT:
      if (dbevent.u.CreateProcessInfo.hProcess)
         CloseHandle(dbevent.u.CreateProcessInfo.hProcess);
      if (dbevent.u.CreateProcessInfo.hThread)
         CloseHandle(dbevent.u.CreateProcessInfo.hThread);
      if (dbevent.u.CreateProcessInfo.hFile)
         CloseHandle(dbevent.u.CreateProcessInfo.hFile);
      break;
   case EXCEPTION_DEBUG_EVENT:
   case EXIT_THREAD_DEBUG_EVENT:
   case EXIT_PROCESS_DEBUG_EVENT:
   case UNLOAD_DLL_DEBUG_EVENT:
   case OUTPUT_DEBUG_STRING_EVENT:
   case RIP_EVENT: 
      // do nothing
      break;
   case LOAD_DLL_DEBUG_EVENT:
      if (dbevent.u.LoadDll.hFile)
         CloseHandle(dbevent.u.LoadDll.hFile);
      break;
   }
}

// the process is going away...
LRESULT CDebugInfoView::OnClose()
{
   // ensure that no threads can be created while we are clearing up
   m_whileDieing = true;
   // tell the other threads to die
   SetEvent(m_hDie);
   // wait for the debug stream thread to go away
   // use this loop to allow the UI to update
   while (MsgWaitForMultipleObjects(1, &m_hThread, FALSE, INFINITE, QS_ALLEVENTS) != WAIT_OBJECT_0)
   {
      MSG msg;
      while (PeekMessage(&msg, this->m_hWnd, 0, 0, PM_REMOVE))
      {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
   }
   CloseHandle(m_hThread);
   KillDebugThreads();
   CloseHandle(m_hDie);
   m_whileDieing = false;
   ClearControls();
   return 0;
}

// search the command line for a particular switch
// if that switch is found, and strValue is not NULL, it is returned
BOOL CDebugInfoView::GetSwitch(LPCTSTR strSwitch, LPTSTR strValue)
{
   LPCTSTR strCmdLine = m_strCommandLine;
   while (*strCmdLine != 0)
   {
      if (*strCmdLine == _T('-') || *strCmdLine == _T('/'))
      {
         strCmdLine++;
         TCHAR strThisSwitch[20];
         int index = 0;
         while (*strCmdLine != _T(' '))
         {
            strThisSwitch[index] = *strCmdLine;
            strCmdLine++;
            index++;
         }
         strThisSwitch[index] = 0;
         if (lstrcmp(strThisSwitch, strSwitch) == 0)
         {
            if (strValue == NULL)
               return TRUE;
            //copy the switch value
            strCmdLine++;
            index = 0;
            while (*strCmdLine != _T(' '))
            {
               strValue[index] = *strCmdLine;
               strCmdLine++;
               index++;
            }
            strValue[index] = 0;
            return TRUE;
         }
      }
      
      strCmdLine++;
   }
   return FALSE;
}

// this is called to clear all the controls on the dialog
LRESULT CDebugInfoView::OnClearAll(WORD, WORD, HWND, BOOL&)
{
   // ensure that no threads can be created while we are clearing up
   m_whileDieing = true;
   // tell all the threads to die (can't avoid the debug stream thread being killed)
   SetEvent(m_hDie);
   // use this loop to allow the UI to update
   while (MsgWaitForMultipleObjects(1, &m_hThread, FALSE, INFINITE, QS_ALLEVENTS) != WAIT_OBJECT_0)
   {
      MSG msg;
      while (PeekMessage(&msg, this->m_hWnd, 0, 0, PM_REMOVE))
      {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
   }

   // kill the processes that we have attached to
   KillDebugThreads();

   // reset the event so that we can allow new threads to run
   ResetEvent(m_hDie);

   // clear the tree view, and disable the menu item to save data and clear data
   ClearControls();

   // update the UI elements
   m_pParent->UIEnable(ID_FILE_SAVERESULTS, FALSE);
   m_pParent->UIEnable(ID_VIEW_CLEARALL, FALSE);
   // update the status bar
   ::PostMessage(GetParent(), WM_UPDATESB, 0, 0);

   // clear list box
   CListBox lb = GetDlgItem(IDC_MSGS);
   lb.ResetContent();

   // clear static control
   CStatic stat = GetDlgItem(IDC_PROCESSPATH);
   stat.SetWindowText(_T(""));

   // need to start the debug stream thread again...
   DWORD dwThreadId;
   m_hThread = CreateThread(NULL, 0, PollDebugStream, this, 0, &dwThreadId);
   m_whileDieing = false;
   return 0;
}

// clean up the controls
void CDebugInfoView::ClearControls()
{
	CTreeViewCtrl tv;
   tv = GetDlgItem(IDC_TREEDATA);

   int count;
   count = tv.GetCount();
   HTREEITEM hItem;
   hItem = tv.GetNextItem(TVGN_ROOT, 0);
   // go through each item in the tree view and release its bucket
   while (count--)
   {
      TVITEM item;
      item.hItem = hItem;
      item.mask = TVIF_PARAM | TVIF_HANDLE;
      tv.GetItem(&item);

      // release the data associated with the tree view item
      CBucket* pBucket = reinterpret_cast<CBucket*>(item.lParam);
      if (pBucket)
         delete pBucket;
      hItem = tv.GetNextItem(hItem, TVGN_NEXT);
   }
   tv.DeleteItem(TVI_ROOT);
}

// start a new process under a debugger thread
void CDebugInfoView::StartProcess(LPCTSTR strFileName)
{
   if (m_whileDieing) 
      // if we are cleaning up the threads then prevent a new one being created
      return;

   EXEData* pData = new EXEData;
   pData->pThis = this;
   
   lstrcpy(pData->strEXE, strFileName);
   HANDLE hThread;
   DWORD dwThreadId;
   hThread = CreateThread(NULL, 0, PollDebuggerEvents, pData, 0, &dwThreadId);
   if (hThread == NULL)
   {
      delete pData;
   }
}

// if we are attached to other processes we must wait for them to go away
void CDebugInfoView::KillDebugThreads()
{
   m_threadArrayCS.Lock();
   // wait for all of the threads to die, I hope there are not more than 
   // MAXIMUM_WAIT_OBJECTS-1 threads, but there maybe. Note the -1, which
   // is a requirement of MsgWaitForMultipleObjects
   if (m_hDebugThreads.size() > 0)
   {
      HANDLE pThreads[MAXIMUM_WAIT_OBJECTS-1];

      DWORD dwSize = m_hDebugThreads.size();
      if (dwSize > MAXIMUM_WAIT_OBJECTS-1) 
         dwSize = MAXIMUM_WAIT_OBJECTS-1;

      DWORD dw;
      for (dw = 0; dw < dwSize; dw++)
      {
         // get the handle out of the vector and remove the value
         std::vector<ThreadData>::iterator it;
         it = m_hDebugThreads.begin();
         ThreadData& td = *it;
         pThreads[dw] = td.hThread;
         m_hDebugThreads.erase(it);
      }

      while (MsgWaitForMultipleObjects(dwSize, pThreads, FALSE, INFINITE, QS_ALLEVENTS) >= WAIT_OBJECT_0 + dwSize)
      {
         MSG msg;
         while(PeekMessage(&msg, this->m_hWnd, 0, 0, PM_REMOVE))
         {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
         }
      }

      for (dw = 0; dw < dwSize; dw++)
         CloseHandle(pThreads[dw]);
   }
   m_threadArrayCS.Unlock();
}

// fill the controls with data saved from another run
void CDebugInfoView::LoadResults(LPCTSTR strFileName)
{
   // the data type is ANSI
   CBucketFile hFile("DebugInfo", 1);
   hFile.Open(strFileName);
   if (hFile.IsOpen())
   {
      BOOL bHandled = TRUE;
      OnClearAll(0, 0, 0, bHandled);
      while (true)
      {
         CBucket* pBucket = new CBucket(0);
         if (!hFile.FillBucket(pBucket))
         {
            delete pBucket;
            break;
         }

         TCHAR sz[20];
         wsprintf(sz, _T("%ld"), pBucket->m_pid);

	      CTreeViewCtrl tv;
         tv = GetDlgItem(IDC_TREEDATA);
         TVINSERTSTRUCT tvis = {TVI_ROOT, TVI_LAST, {TVIF_TEXT|TVIF_PARAM|TVIF_IMAGE, 0, 0, 0, sz, 0, 0, 0, 0, reinterpret_cast<LPARAM>(pBucket)}};
         tv.InsertItem(&tvis);
      }
   }
}

// save these results
void CDebugInfoView::SaveResults(LPCTSTR strFileName)
{
	CTreeViewCtrl tv;
   tv = GetDlgItem(IDC_TREEDATA);
   int count;

   count = tv.GetCount();
   // the data type is ANSI
   CBucketFile hFile("DebugInfo", 1);
   hFile.Create(strFileName);
   if (hFile.IsOpen())
   {
      HTREEITEM hItem = NULL;
      hItem = tv.GetNextItem(TVGN_ROOT, 0);
      while (count--)
      {
         TVITEM item;
         memset(&item, 0, sizeof(TVITEM));
         item.hItem = hItem;
         item.mask = TVIF_PARAM | TVIF_HANDLE;
         tv.GetItem(&item);
         hFile.SaveBucket(reinterpret_cast<CBucket*>(item.lParam));
         hItem = tv.GetNextItem(hItem, TVGN_NEXT);
      }
   }
}
