// DebugInfoView.h : interface of the CDebugInfoView class
//
/////////////////////////////////////////////////////////////////////////////

// (c) 2000 Richard Grimes, iDevResource.com Ltd
// you are free to use this code as long as your copyright message
// acknowledges iDevResource.com Ltd

#if !defined(AFX_DEBUGINFOVIEW_H__1FF3DEEC_D4D4_11D3_8A65_00104BDC35E0__INCLUDED_)
#define AFX_DEBUGINFOVIEW_H__1FF3DEEC_D4D4_11D3_8A65_00104BDC35E0__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <vector>
#include "Bucket.h"

#define TIMEOUT 100
#define TAB 3
#ifndef NAME_MAX
   #define NAME_MAX 100
#endif

// this is the class that shows the UI
// it also has the 'document'
class CDebugInfoView : public CDialogImpl<CDebugInfoView>
{
public:
	enum { IDD = IDD_DEBUGINFO_FORM };

   CDebugInfoView(CUpdateUIBase* pParent) 
      : m_dwTimeout(50), m_whileDieing(false), m_pParent(pParent)
   {}

	BOOL PreTranslateMessage(MSG* pMsg)
	{
		return IsDialogMessage(pMsg);
	}

   // Should use BEGIN_MSG_MAP_EX() for the message map because I am using crackers
   // but this screws up classview. Instead, declared the functions by hand
	BOOL m_bATL3MsgHandled; 
	BOOL IsMsgHandled() const 
	{ 
		return m_bATL3MsgHandled; 
	} 
	void SetMsgHandled(BOOL bHandled) 
	{ 
		m_bATL3MsgHandled = bHandled; 
	} 

	BEGIN_MSG_MAP(CDebugInfoView)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_CLOSE(OnClose)
      MSG_WM_SIZE(OnSize)
      MSG_WM_GETMINMAXINFO(OnGetMinMax)
      MESSAGE_HANDLER(WM_NEWDATA, OnNewData)
		COMMAND_ID_HANDLER(ID_VIEW_CLEARALL, OnClearAll)
      NOTIFY_HANDLER(IDC_TREEDATA, TVN_SELCHANGED, OnSelectedPID)
	END_MSG_MAP()

	LRESULT OnInitDialog(HWND, LPARAM);
   LRESULT OnGetMinMax(LPMINMAXINFO lpMinMax);
	LRESULT OnSelectedPID(int idCtrl, LPNMHDR pnmh, BOOL& bHandled);
	LRESULT OnSize(UINT flags, CSize size);
	LRESULT OnNewData(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
	LRESULT OnClose();
   LRESULT OnClearAll(WORD, WORD, HWND, BOOL&);

   void StartProcess(LPCTSTR strFileName);
   void LoadResults(LPCTSTR strFileName);
   void SaveResults(LPCTSTR strFileName);

private:
   LPTSTR SplitString(LPTSTR str, CBucket* pBucket);
   void SizeListbox(HWND hWnd, LPTSTR str);
   void KillDebugThreads();
	void ClearControls();
   BOOL GetSwitch(LPCTSTR strSwitch, LPTSTR strValue);
   static DWORD WINAPI PollDebugStream(LPVOID pv);
   static DWORD WINAPI PollDebuggerEvents(LPVOID pv);
   static void CleanUpEventResources(DEBUG_EVENT dbevent);

private:
   // information about an EXE needed by the debugging routines
   typedef struct _EXEData
   {
      CDebugInfoView* pThis;
      TCHAR strEXE[NAME_MAX];
   } EXEData;
   // information about DLLs loaded by a debugged process
   typedef struct _DllData
   {
      LPVOID lpBase;
      TCHAR strDLL[MAX_PATH];
   } DLLData;
   // information about a thread that is attached as a debugger to a process
   typedef struct _ThreadData
   {
      DWORD pid;
      HANDLE hThread;
      TCHAR str[MAX_PATH];
   } ThreadData;

private:   
   HANDLE m_hDie;
   HANDLE m_hThread;
   DWORD m_dwTimeout;
   CString m_strCommandLine;
   std::vector<ThreadData> m_hDebugThreads;
   CComAutoCriticalSection m_threadArrayCS;
   CImageList m_imageList;
   CUpdateUIBase* m_pParent;
   bool m_whileDieing;
};


/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_DEBUGINFOVIEW_H__1FF3DEEC_D4D4_11D3_8A65_00104BDC35E0__INCLUDED_)
